"use strict";

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _rvjsDom = require("rvjs-dom");

var _rvjsDomElement = require("rvjs-dom-element");

var _rvjsDomElement2 = _interopRequireDefault(_rvjsDomElement);

var _measureScrollbar = require("./measure-scrollbar");

var _measureScrollbar2 = _interopRequireDefault(_measureScrollbar);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var Noop = function Noop() {};

var KeyCodes = {
	// '*': 'navigate',
	13: 'enter',
	27: 'escape',
	32: 'space',
	37: 'left', // leftRight
	38: 'up', // upDown
	39: 'right',
	40: 'down'
};

var KeyCodesAll = Object.keys(KeyCodes).map(function (code) {
	return KeyCodes[code];
});

var EvnAlias = {
	leftRight: ['left', 'right'],
	upDown: ['up', 'down'],
	navigate: KeyCodesAll
};

var IgnoreTarget = ['TEXTAREA', 'INPUT', 'SELECT', 'A', 'BUTTON'];

var IsInit = false;
var IsInitComplete = false;
var Reg = {};
var Initable = {};
var Keys = [];
var InitKeys = [];
var HtmlLayer = null,
    HtmlBackground = null,
    Html = null,
    HtmlBody = null;
var Layers = [];
var IsOpen = false;
var IsPreload = false;
var _Off = Noop;
var WinEvents = {};

var Props = {
	className: 'shadow-layer',
	backgroundClassName: 'shadow-layer-background',
	wrapClassName: 'shadow-layer-wrap',
	escape: true,
	fixed: 'body'
};

/**
 * @return {boolean}
 */
function IsContentEditable(e) {
	var src = e.target;
	if (src && src.nodeName) {
		if (IgnoreTarget.indexOf(src.nodeName.toUpperCase()) > -1) {
			return true;
		}

		while (src) {
			if (src.contentEditable === 'true') return true;else if (src.contentEditable === 'inherit') src = src.parentNode;else break;
		}
	}

	return false;
}

function Call(name, func) {
	for (var _len = arguments.length, evn = Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
		evn[_key - 2] = arguments[_key];
	}

	var layer = Reg[name] || {};

	if (typeof layer[func] !== "function") {
		return func === 'reload' ? Call.apply(undefined, [name, 'load'].concat(_toConsumableArray(evn))) : void 0;
	} else if (!evn.length) {
		evn = [HtmlLayer];
	}

	return layer[func].apply(layer, evn);
}

function Dispatch(event) {
	var app = Props.app || false;
	if (app && typeof app.dispatch === "function") {
		app.dispatch(event);
	}
}

function LayerOpen() {
	if (!IsOpen) {
		IsOpen = true;
		HtmlBackground.show(350);
		HtmlLayer.show(350);
		Html.className('+layer-open');
		Props.fixed && _rvjsDom.Element.css(Props.fixed, { paddingRight: (0, _measureScrollbar2.default)() });
		_rvjsDom.Element.css('body', { overflow: 'hidden' });

		Dispatch({
			type: "layer_open"
		});

		_Off();

		_rvjsDom.Evn.on("keyup", function (e) {
			var evn = WinEvents[Layer.name || ''],
			    code = e.keyCode || 0;

			if (evn && KeyCodes[code]) {
				code = KeyCodes[code];
				evn[code] && !IsContentEditable(e) && evn[code](code, e) !== false && e.preventDefault();
			}
		}, function (off) {
			_Off = function Off() {
				off();
				_Off = Noop;
			};
		});
	}
}

function LayerClose() {
	for (var i = Layers.length - 1; i >= 0; i--) {
		if (Layers[i].close() === false) {
			return;
		}
	}

	if (IsOpen) {
		IsOpen = false;
		HtmlLayer.hide(350);
		HtmlBackground.hide(350, function () {
			HtmlBody.data('layer-name', null);
			Html.className('-layer-open');
			Props.fixed && _rvjsDom.Element.css(Props.fixed, { paddingRight: '' });
			_rvjsDom.Element.css('body', { overflow: '' });
			Dispatch({
				type: "layer_close"
			});
			_Off();
		});
	}
}

/**
 * @return {boolean}
 */
function CloseItem(layer, trigger) {
	if (trigger && Call(layer.name, "blur", layer.element, layer.result) === false) {
		return false;
	}

	Dispatch({
		type: "layer_unload",
		name: layer.name,
		layer: layer.result
	});

	Call(layer.name, "close", layer.element, layer.result);
	RecalculateIndexes(layer.index);
	if (typeof Props.garbage === "function") {
		try {
			Props.garbage(layer.element, layer.name);
		} catch (e) {}
	}
	HtmlLayer.element.removeChild(layer.element);

	if (IsPreload) {
		return true;
	}

	if (Layers.length < 1) {
		LayerClose();
	} else if (!trigger) {
		// change depth because user call
		Depth();
	}

	return true;
}

function Depth() {
	for (var i = Layers.length - 1, cur = i, depth; i >= 0; i--) {
		depth = cur - i;
		Layers[i].element.className = Props.wrapClassName + (depth < 4 ? " layer-index-" + (depth + 1) : "") + (depth < 1 ? "" : depth < 3 ? " layer-blur" : " layer-hidden");
	}

	var last = Layers.length - 1;
	if (last > -1) {
		HtmlBody.data('layer-name', Layers[last].name);
	}
}

function RecalculateIndexes(index) {
	Layers.splice(index, 1);
	for (var i = 0; i < Layers.length; i++) {
		Layers[i].index = i;
	}
}

function Assign(data, rst) {
	if ((typeof rst === "undefined" ? "undefined" : _typeof(rst)) === 'object' && rst !== null) {
		Object.keys(rst).forEach(function (name) {
			if (name !== 'id' && name !== 'close') {
				data[name] = rst[name];
			}
		});
	}

	return data;
}

/**
 * @return {function|boolean}
 */
function Preload(name, data, show) {
	if ((typeof name === "undefined" ? "undefined" : _typeof(name)) === 'object') {
		Object.assign(data, name);
		name = data.name || '';
	}

	if (!Layer.loaded(name)) {
		throw new Error("Layer controller '" + name + "' not found");
	}

	// check global init
	if (!IsInitComplete) {
		Layer.init();
	}
	if (!IsInit) {
		throw new Error("Global html container is not loaded");
	}

	// init layer
	if (InitKeys.indexOf(name) < 0) {
		InitKeys.push(name);
		Call(name, "init", InitLayer(name, Initable[name]));
	}

	// crate auto id
	if (!data.id) {
		data.id = 'layer_' + Math.floor(Math.random() * 10000000);
	}

	IsPreload = true;

	var cur = null,
	    // current layer
	reload = false,
	    // reload layer
	rst = void 0,
	    stop = function stop() {
		IsPreload = false;
		return IsPreload;
	};

	// show layer, add next
	if (show) {
		for (var i = 0, length = Layer.length - 1, layer; i < length; i++) {
			layer = Layers[i];
			if (layer.show && layer.id === data.id) {
				cur = layer;
				reload = true;
				break;
			}
		}
	} else {

		// close parent layers
		for (var _i = Layers.length - 1, _layer; _i >= 0; _i--) {
			_layer = Layers[_i];
			if (!_layer.show) {
				cur = _layer;
			} else if (!_layer.close()) {
				return stop();
			}
		}

		if (cur) {
			if (cur.name === name) {
				reload = true;
			} else if (!cur.close()) {
				return stop();
			}
		}
	}

	// IsPreload = false
	stop();

	if (reload) {

		rst = Call(cur.name, "reload", cur.element, Assign(cur.result, data));
		if (rst !== false) {

			cur.result = Assign(cur.result, rst);

			// move layer to top
			if (cur.index !== Layers.length - 1) {
				RecalculateIndexes(cur.index);
				cur.index = Layers.length;
				Layers.push(cur);
			}
		}
	} else {

		cur = {
			id: data.id,
			name: name,
			show: show,
			index: Layers.length,
			element: _rvjsDom.Element.create({ parent: HtmlLayer.element, data: { closable: 'true' } })
		};

		cur.close = data.close = CloseItem.bind(Layer, cur, true);

		Layer.init();
		rst = Call(cur.name, "load", cur.element, data);

		if (rst !== false) {
			LayerOpen();
			Layers.push(cur);
		}
	}

	if (rst !== false) {

		// set layer result for call

		cur.result = Assign(data, rst);

		// move levels, z-index, view mode

		Depth();

		Dispatch({
			type: "layer_load",
			name: cur.name,
			reload: reload,
			layer: cur.result
		});

		return CloseItem.bind(Layer, cur, false);
	} else if (Layers.length < 1) {
		LayerClose();
	}

	return false;
}

function CreateElement(className) {
	return new _rvjsDomElement2.default({
		name: "div.fade." + className,
		style: { display: 'none' },
		events: {
			click: function click(e) {
				var target = e.target;
				target && target.dataset.closable === 'true' && Layer.hide();
			}
		},
		data: { closable: true },
		parent: document.body
	});
}

/**
 * @return {boolean}
 */
function IsFunction(value) {
	return typeof value === 'function';
}

function InitLayer(name, data) {
	WinEvents[name] = {};
	var CopyOf = {},
	    Evn = WinEvents[name];

	Object.keys(KeyCodes).forEach(function (code) {
		name = KeyCodes[code];
		if (IsFunction(data[name])) {
			Evn[name] = data[name];
		}
	});

	Object.keys(EvnAlias).forEach(function (code) {
		if (IsFunction(data[code])) {
			EvnAlias[code].forEach(function (name) {
				if (!Evn[name]) {
					Evn[name] = data[code];
				}
			});
		}
	});

	if (!Evn.escape && (data.escape === true || Props.escape && data.escape !== false)) {
		Evn.escape = Layer.hide;
	}

	// copy data object, ignore event flag and callback

	Object.keys(data).forEach(function (key) {
		if (KeyCodesAll.indexOf(key) < 0 && !EvnAlias[key]) {
			CopyOf[key] = data[key];
		}
	});

	return CopyOf;
}

function InitHtml() {
	var styles = {},
	    layer = '.' + Props.className,
	    background = '.' + Props.backgroundClassName,
	    wrap = '.' + Props.wrapClassName,
	    scroll = (0, _measureScrollbar2.default)();

	styles[layer + ', ' + background] = 'position:fixed;top:0;left:0;bottom:0;width:100%;height:100%;min-height:100vh;';
	styles[background] = 'z-index:1000;background:rgba(0,0,0,.7);';
	styles[layer] = 'z-index:1100;overflow:auto;';
	styles[wrap] = 'position:absolute;top:0;left:0;width:100%;z-index:100;';
	styles[(wrap += '.layer-index-') + 4] = 'z-index:100;';
	styles[wrap + 3] = 'z-index:200;';
	styles[wrap + 2] = 'z-index:300;';
	styles[wrap + 1] = 'z-index:400;';
	styles[':root'] = '--scroll-bar-size:' + scroll + 'px';

	try {
		(0, _rvjsDom.StyleSheets)(styles);
	} catch (e) {}

	Html = new _rvjsDomElement2.default(document.documentElement);
	HtmlBody = new _rvjsDomElement2.default(document.body);
	HtmlBackground = CreateElement(Props.backgroundClassName);
	HtmlLayer = CreateElement(Props.className);
	IsInit = true;
}

var Layer = {
	get name() {
		var index = Layers.length;
		return index > 0 && IsOpen ? Layers[index - 1].name : null;
	},

	get isOpened() {
		return IsOpen && Layers.length > 0;
	},

	init: function init() {
		var props = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

		if (!IsInitComplete) {
			IsInitComplete = true;
			if ((typeof props === "undefined" ? "undefined" : _typeof(props)) === 'object' && props !== null) {
				Object.assign(Props, props);
			}

			(0, _rvjsDom.Ready)(InitHtml);
		}

		return Layer;
	},
	register: function register(name, object, data) {
		if (Keys.indexOf(name) < 0) {
			Keys.push(name);
			Reg[name] = typeof object === 'function' ? { load: object } : object;
			Initable[name] = data || {};
		}
		return Layer;
	},
	loaded: function loaded(name) {
		return Keys.indexOf(name) > -1;
	},
	open: function open(name, data) {
		return Preload(name, data || {}, false);
	},
	show: function show(name, data) {
		return Preload(name, data || {}, true);
	},
	hide: function hide() {
		var id = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

		var cur = null;

		for (var i = Layers.length - 1, layer; i >= 0; i--) {
			layer = Layers[i];
			if (!id || layer.id === id) {
				cur = layer;
				break;
			}
		}

		if (!cur) {
			return null;
		}

		if (!cur.show) {
			return LayerClose();
		}

		if (!cur.close()) {
			return false;
		}

		Layers.length > 0 && Depth();
		return true;
	},
	close: function close() {
		return LayerClose();
	}
};

exports.default = Layer;